from __future__ import annotations
from typing import Any, Callable, Dict

from smarter_client.domain import Device, SmarterClient
from smarter_client.managed_devices.base import BaseDevice


class SmarterKettleV3(BaseDevice):
    device: Device = None
    user_id: str = None
    _status_handler: Any = None

    @classmethod
    def from_id(cls, client: SmarterClient, device_id: str, user_id: str) -> SmarterKettleV3:
        device = Device.from_id(client, device_id)
        device.fetch()
        return cls(device, user_id)

    @classmethod
    def from_device(cls, device: Device, user_id: str) -> SmarterKettleV3:
        device.fetch()
        return cls(device, user_id)

    def __init__(self, device: Device, user_id: str):
        super().__init__(device, 'Smarter Kettle V3', 'kettle', user_id)

    def _on_event(self, event):
        if 'status' not in event.get('path', []):
            return

        if self._status_handler:
            self._status_handler(self.device.status)

    def _send_command(self, command: str, value: Any):
        self.device.commands.get(command).execute(self.user_id, value)

    @property
    def status(self):
        return self.device.status

    @property
    def settings(self):
        return self.device.settings

    def watch_status(self, handler: Callable):
        self._status_handler = handler
        self.device.watch(self._on_event)

    def dispose(self):
        self.device.unwatch()

    def add_alarm(self, value: int):
        self._send_command('add_alarm', value)

    def calibrate_weight_sensor(self, value: int):
        self._send_command('calibrate_weight_sensor', value)

    def change_alarm(self, value: int):
        self._send_command('change_alarm', value)

    def remove_alarm(self, value: int):
        self._send_command('remove_alarm', value)

    def reset_settings(self, value: int):
        self._send_command('reset_settings', value)

    def resync_database(self, value: int):
        self._send_command('resync_database', value)

    def send_notification(self, value: int):
        self._send_command('send_notification', value)

    def send_ping(self, value: int):
        self._send_command('send_ping', value)

    def server_restart(self, value: int):
        self._send_command('server_restart', value)

    def set_boil_temperature(self, value: int):
        self._send_command('set_boil_temperature', value)

    def set_formula_mode_enable(self, value: bool):
        self._send_command('set_formula_mode_enable', value)

    def set_formula_mode_temperature(self, value: int):
        self._send_command('set_formula_mode_temperature', value)

    def set_handle_right_side(self, value: bool):
        self._send_command('set_handle_right_side', value)

    def set_keep_warm_time(self, value: int):
        self._send_command('set_keep_warm_time', value)

    def set_manual_boil_temperature(self, value: int):
        self._send_command('set_manual_boil_temperature', value)

    def set_manual_formula_mode_enable(self, value: bool):
        self._send_command('set_manual_formula_mode_enable', value)

    def set_manual_formula_mode_temperature(self, value: int):
        self._send_command('set_manual_formula_mode_temperature', value)

    def set_manual_keep_warm_time(self, value: int):
        self._send_command('set_manual_keep_warm_time', value)

    def set_options(self, value: int):
        self._send_command('set_options', value)

    def set_region(self, value: int):
        self._send_command('set_region', value)

    def set_user(self, value: Dict[str, Any]):
        self._send_command('set_user', value)

    def start_auto_boil(self):
        # Placeholder value, since no example provided
        self._send_command('start_auto_boil', 0)

    def start_boil(self, value: bool):
        self._send_command('start_boil', value)

    def stats_update(self, value: int):
        self._send_command('stats_update', value)

    def stop_boil(self, value: int):
        self._send_command('stop_boil', value)

    def turn_off_wifi(self, value: int):
        self._send_command('turn_off_wifi', value)




"""The `version` module holds the version information for Smarter Kettle Client."""
from __future__ import annotations as _annotations

__all__ = 'VERSION', 'version_info'

VERSION = '0.1.9'
"""The version of Smarter Kettle Client."""


def version_short() -> str:
    """Return the `major.minor` part of version.

    It returns '2.1' if version is '2.1.1'.
    """
    return '.'.join(VERSION.split('.')[:2])


def version_info() -> str:
    """Return complete version information for package and its dependencies."""
    import platform
    import sys
    from pathlib import Path

    info = {
        'version': VERSION,
        'install path': Path(__file__).resolve().parent,
        'python version': sys.version,
        'platform': platform.platform(),
    }
    return '\n'.join('{:>30} {}'.format(k + ':', str(v).replace('\n', ' ')) for k, v in info.items())



"""
Module containing wrappers for specific devices
"""
from smarter_client.domain import Network, Device
from smarter_client.managed_devices.kettle_v3 import SmarterKettleV3


def load_from_network(network: Network, user_id: str) -> list[Device]:
    """
    Get devices from a given network
    """
    return (
        wrapper for wrapper in
        (
            get_device_wrapper(device, user_id)
            for device in network.associated_devices
        )
        if wrapper is not None
    )


def get_device_wrapper(device: Device, user_id: str):
    """
    Get a device wrapper for a specific user by inferring the correct type
    from the device properties
    """
    device.fetch()
    model = device.status.get('device_model')
    match model:
        case 'SMKET01':
            return SmarterKettleV3.from_device(device, user_id)
        case _:
            print(f'Unknown device model: {model}')



"""
Module contains implementation of Smarter Firebase API
"""
from __future__ import annotations
from typing import Callable
from pyrebase.pyrebase import Stream
import pyrebase

from smarter_client.domain.decorators.session import refreshsession
from .models import LoginSession
from .._consts import API_KEY


class SmarterClient:  # pragma: no cover
    """
    A client for interacting with the Smarter API.
    """
    session: LoginSession

    def __init__(self):
        config = {
            "apiKey": API_KEY,
            "authDomain": "smarter-live.firebaseapp.com",
            "databaseURL": "https://smarter-live.firebaseio.com",
            "projectId": "smarter-live",
            "storageBucket": "smarter-live.appspot.com",
            "messagingSenderId": "41919779740"
        }
        app = pyrebase.initialize_app(config)
        self.app = app
        self.token = None

    def sign_in(self, email, password) -> LoginSession:
        auth = self.app.auth()

        user = auth.sign_in_with_email_and_password(email, password)
        self.token = user.get("idToken")
        self.session = LoginSession(user)
        return self.session

    def refresh(self):
        auth = self.app.auth()
        refresh_response = auth.refresh(self.session.refresh_token)
        self.token = refresh_response.get("idToken")
        self.session.update(refresh_response)
        return self.session

    @refreshsession
    def get_user(self, user_id: str):
        database = self.app.database()
        return database.child("users").child(user_id).get(self.token).val()

    @refreshsession
    def get_network(self, network_id: str):
        database = self.app.database()
        return database.child("networks").child(network_id).get(self.token).val()

    @refreshsession
    def get_device(self, device_id: str):
        database = self.app.database()
        return database.child('devices').child(device_id).get(self.token).val()

    @refreshsession
    def get_status(self, device_id: str):
        database = self.app.database()
        return database.child('devices').child(device_id).child('status').get(self.token).val()

    def get_db(self):
        return self.app.database()

    @refreshsession
    def send_command(self, device_id: str, command: str, data: dict):
        database = self.app.database()

        return database \
            .child('devices') \
            .child(device_id) \
            .child('commands') \
            .child(command) \
            .push(data, self.token)

    @refreshsession
    def watch_device_attribute(self, device_id: str, callback) -> Callable:
        database = self.app.database()
        stream = database.child('devices').child(
            device_id).stream(callback, self.token)
        return lambda: stream.close()



from abc import ABCMeta
from smarter_client.domain.models import Device


class BaseDevice(metaclass=ABCMeta):
    device: Device
    friendly_name: str
    type: str
    user_id: str

    def __init__(self, device: Device, friendly_name: str, device_type: str, user_id: str):
        device.fetch()
        self.device = device
        self.friendly_name = friendly_name
        self.user_id = user_id
        self.type = device_type

    @property
    def id(self):
        return self.device.identifier

    @property
    def model(self):
        return self.device.status.get('device_model')

    def firmware_version(self):
        return self.device.status.get('firmware_version')

    def __str__(self):
        return f"Device ID: {self.id}, Device Name: {self.friendly_name} ({self.model}), Device Type: {self.type}"



from copy import deepcopy
from collections import deque


def _tokenize_path(path: str):
    return path.strip('/').split('/') if path != '/' else [None]


def _find_or_create_node(target: dict, path: str) -> tuple[dict, str]:
    *tokens, child = _tokenize_path(path)
    selector = deque(tokens)
    current_node = target
    while len(selector) > 0:
        next_key = selector.popleft()

        if next_key not in current_node:
            current_node[next_key] = {}

        # if len(selector) == 0:
        #     return current_node[next_key]
        # else:
        current_node = current_node[next_key]

    if child is not None and child not in current_node:
        current_node[child] = {}

    return (current_node, child)


def patch_dict(target: dict, path: str, patch: dict) -> dict:
    new_dict = deepcopy(target)
    parent, child = _find_or_create_node(new_dict, path)

    if child is None:
        parent.update(patch)
    else:
        parent[child].update(patch)

    return new_dict


def put_dict(target: dict, path: str, put: dict) -> dict:
    new_dict = deepcopy(target)
    parent, child = _find_or_create_node(new_dict, path)

    if child is None:
        return put
    else:
        parent[child] = put

    return new_dict


def delete_dict(target: dict, path: str) -> dict:
    new_dict = deepcopy(target)
    parent, child = _find_or_create_node(new_dict, path)

    if child is None:
        raise ValueError('Cannot delete root of dict')
    else:
        del parent[child]

    return new_dict

# def xpatch_dict(target: dict, path: patch: dict):
#   for key, value in source.items():
#       if isinstance(value, dict):
#           if key not in target:
#               target[key] = {}
#           patch_dict(target[key], value)
#       if
#       else:
#           target[key] = value
#   return target



from __future__ import annotations


def refreshsession(func):
    def wrapper(self, *args, **kwargs):
        if self.session is None:
            raise ValueError("No session")
        if self.session.expires_in < 30:
            self.refresh()
        return func(self, *args, **kwargs)
    return wrapper



import sys
from unittest.mock import Mock


def pytest_configure(config):
    """
    Allows plugins and conftest files to perform initial configuration.
    This hook is called for every plugin and initial conftest
    file after command line options have been parsed.
    """


def pytest_sessionstart(session):
    """
    Called after the Session object has been created and
    before performing collection and entering the run test loop.
    """
    module = type(sys)('pyrebase')
    module.pyrebase = type(sys)("pyrebase")
    module.pyrebase.Stream = Mock()
    # module.sum = fwlib_sum
    sys.modules['pyrebase'] = module
    sys.modules['pyrebase.pyrebase'] = module.pyrebase


def pytest_sessionfinish(session, exitstatus):
    """
    Called after whole test run finished, right before
    returning the exit status to the system.
    """


def pytest_unconfigure(config):
    """
    called before test process is exited.
    """



from __future__ import annotations

from abc import ABCMeta
from abc import abstractmethod
from typing import Any, Callable, Self
from pprint import pprint

import datetime
from smarter_client.dict_util import delete_dict, patch_dict, put_dict
from . import smarter_client


class BaseEntity(metaclass=ABCMeta):
    identifier: str = None
    _data: dict = None
    client: smarter_client.SmarterClient
    is_stub: bool = True

    # Class methods
    @classmethod
    def from_id(cls, client: smarter_client.SmarterClient, identifier: str) -> Self:
        self = cls(client)
        self.client = client
        self.is_stub = True
        self.identifier = identifier

        return self

    @classmethod
    def from_data(cls, client: smarter_client.SmarterClient, data: dict, identifier: str = None) -> Self:
        self = cls(client)
        self._data = data
        self.identifier = identifier
        self._init_data()

        return self

    def __init__(self, client: smarter_client.SmarterClient):
        self.client = client

    # Public methods
    def fetch(self):
        if self.is_stub:
            self._data = self._fetch()
            self._init_data()
            self.is_stub = False

    # Private methods

    def _get_handler(self, event: dict):
        event_name: str = event.get('event')
        data: dict = event.get('data')

        if event_name == 'patch':
            return patch_dict
        if event_name == 'put' and data is None:
            return delete_dict
        if event_name == 'put':
            return put_dict

        return lambda: None

    def _on_event(self, event):
        event_name: str = event.get('event')
        path: str = event.get('path')
        data: dict = event.get('data')

        if event_name not in ('put', 'patch'):
            print(f'Unexpected event: {event_name}')
            pprint(event)
            return

        handler = self._get_handler(event)

        try:
            self._data = handler(self._data, path, data)
            self._init_data()
        except BaseException as e:
            print(f'Error handling update: {e}')
            pprint((event_name, path, data))

    @abstractmethod
    def _fetch(self) -> dict:
        pass

    @abstractmethod
    def _init_data(self) -> None:
        pass

    def __str__(self):
        return f'{self.__class__.__name__}({self.identifier})'


# </BaseEntiy>


class Commands(BaseEntity, dict[str, 'Command']):
    device: Device

    # Class methods
    @classmethod
    def from_data(cls, client: smarter_client.SmarterClient, data: dict, device: Device) -> Self:
        self = Commands(client, device)
        self._data = data
        self._init_data()

        return self

    @classmethod
    def from_id(cls) -> Self:
        raise RuntimeError('Cannot create Commands from id')

    def __init__(self, client: smarter_client.SmarterClient, device: Device):
        super().__init__(client)
        self.device: Device = device

    # Private methods
    def _init_data(self) -> None:
        self._build_commands()

    def _build_commands(self):
        super().clear()
        super().update({
            key: Command.from_data(self.client, value, key, self.device)
            for key, value
            in self._data.items()
        })

    def _fetch(self) -> dict:  # pragma: no cover
        raise NotImplementedError()


# </Commands>


class CommandInstance(BaseEntity):
    command: Command = None
    device: Device = None
    value: dict[str, Any] = None
    state: str = None
    response: dict[str, Any] = None
    user_id: str = None

    # Class Methods
    def __init__(self, client: smarter_client.SmarterClient):
        super().__init__(client)

    def _init_data(self) -> None:
        self.user_id = self._data.get('user_id')
        self.value = self._data.get('value')
        self.state = self._data.get('state')
        self.response = self._data.get('response')

    def _fetch(self) -> dict:  # pragma: no cover
        raise NotImplementedError()

    @classmethod
    def from_data(cls,
                  client: smarter_client.SmarterClient,
                  data: dict,
                  identifier: str,
                  command: Command,
                  device: 'Device') -> Self:
        self = CommandInstance(client)
        self.command = command
        self.identifier = identifier
        self._data = data
        self.device = device
        self._init_data()

        return self

# </CommandInstance>


class Command(BaseEntity):
    name: str = None
    device: Device = None
    instances: dict[str, CommandInstance] = {}
    example: dict = None
    # Class methods

    @classmethod
    def from_data(cls,
                  client: smarter_client.SmarterClient,
                  data: dict,
                  name: str,
                  device: Device) -> Self:
        self = Command(client)
        self._data = data
        self.identifier = name
        self.device = device
        self._init_data()

        return self

    def __init__(self, client: smarter_client.SmarterClient):
        super().__init__(client)

    def execute(self, user_id: str, value: any):
        return self.client.send_command(self.device.identifier, self.name, {"user_id": user_id, "value": value})

    def _init_data(self) -> None:
        self.name = self.identifier
        self.example = self._data.get('example')
        self.instances = {
            key: CommandInstance.from_data(
                self.client, value, key, self, self.device)
            for key, value
            in self._data.items()
            if key != 'example'
        }

    def _fetch(self) -> dict:  # pragma: no cover
        raise NotImplementedError()
# </Command>


class Device(BaseEntity):
    commands: Commands = None
    settings: Settings = None
    status: Status = None
    _stream_close: Callable = None

    def __init__(self, client: smarter_client.SmarterClient):
        super().__init__(client)

    # Public methods
    def watch(self, callback=lambda: None):
        if self._stream_close is not None:
            raise RuntimeError(
                'Already watching device. Call unwatch() first. Support for multiple callbacks may be implemented in a later version')

        def on_data(event):
            self._on_event(event)
            callback(event)
        self._stream_close = self.client.watch_device_attribute(
            self.identifier, on_data)

    def unwatch(self):
        if self._stream_close is not None:
            self._stream_close()
            self._stream_close = None

    # Private methods
    def _init_data(self):
        self.commands = Commands.from_data(
            self.client, self._data.get('commands'), self)

        self.settings = Settings.from_data(
            self.client, self._data.get('settings'))

        # Status(client: smarter_client.SmarterClient, data.get('status'))
        self.status = Status.from_data(
            self.client, self._data.get('status'), self)

    def _fetch(self) -> dict:
        return self.client.get_device(self.identifier)

# </Device>


class LoginSession:
    kind: str = None
    local_id: str = None
    email: str = None
    display_name: str = None
    id_token: str = None
    registered: bool = None
    refresh_token: str = None
    session_duration: int = None
    expires_at: datetime.datetime = None

    def __init__(self, data: dict):
        self.kind = data.get('kind')
        self.local_id = data.get('localId')
        self.email = data.get('email')
        self.display_name = data.get('displayName')
        self.id_token = data.get('idToken')
        self.registered = data.get('registered')
        self.refresh_token = data.get('refreshToken')
        self.session_duration = data.get('expiresIn')

        self.expires_at = self._get_expiration_datetime(data)

    def _get_expiration_datetime(self, data: dict):
        return datetime.datetime.now() + datetime.timedelta(0, int(data.get('expiresIn')))

    def is_expired(self) -> bool:
        """
        Returns True if the session has expired.
        """
        return self.expires_at <= datetime.datetime.now()

    @property
    def expires_in(self) -> int:
        """
        Returns the number of seconds until the session expires.
        """
        return (self.expires_at - datetime.datetime.now()).total_seconds()

    def update(self, data: dict):
        self.id_token = data.get('idToken')
        self.refresh_token = data.get('refreshToken')
        self.local_id = data.get('userId')
        self.expires_at = self._get_expiration_datetime(data)


# </LoginSession>

class Network(BaseEntity):
    access_tokens_fcm: dict[str, str] = None
    associated_devices: list[Device] = None
    name: str = None
    owner: User = None

    def __init__(self, client):
        super().__init__(client)
        self.access_tokens_fcm = None
        self.associated_devices = None
        self.name = None
        self.owner = None

    def _fetch(self) -> dict:
        return self.client.get_network(self.identifier)

    def _init_data(self) -> None:
        self.access_tokens_fcm = self._data.get('access_tokens_fcm')
        self.associated_devices = [Device.from_id(self.client,
                                                  key) for key in self._data.get('associated_devices')]

        self.name = self._data.get('name')
        self.owner = User.from_id(self.client, self._data.get('owner'))

# </Network>


class Settings(BaseEntity):
    def __init__(self, client: smarter_client.SmarterClient):
        super().__init__(client)
        self.identifier = '/'
        self.network: Network = None
        self.network_ssid: str = None

    def _fetch(self) -> dict:
        return dict()

    def _init_data(self) -> None:
        self.network = Network.from_id(self.client, self._data.get('network'))
        self.network_ssid = self._data.get('network_ssid')

# </Settings>


class Status(BaseEntity, dict):
    device: Device

    # Class methods
    @classmethod
    def from_data(cls, client: smarter_client.SmarterClient, data: dict, device: Device) -> Self:
        self = super().from_data(client, data, f'{device.identifier}/status')
        self.device = device

        return self

    def __init__(self, client: smarter_client.SmarterClient):
        super().__init__(client)

    def _init_data(self) -> None:
        self.clear()
        self.update(self._data)

    def _fetch(self) -> dict:
        self.client.get_status(self.device.identifier)

# </Status>


class User(BaseEntity):
    email: str = None
    accepted: int = None
    first_name: str = None
    last_name: str = None
    location_accepted: int = None
    networks_index: dict[str, str] = None
    temperature_unit: int = None
    networks: dict[str, Network] = dict()

    def __init__(self, client):
        super().__init__(client)

    def _init_data(self) -> None:
        self.accepted: int = datetime.datetime.fromtimestamp(
            self._data.get('accepted')/1000.0)
        self.email: str = self._data.get('email')
        self.first_name: str = self._data.get('first_name')
        self.last_name: str = self._data.get('last_name')
        self.location_accepted: int = self._data.get('locationAccepted')
        self.networks_index: dict = self._data.get('networks_index')
        self.temperature_unit: int = self._data.get('temperature_unit')

        self.networks = {value: Network.from_id(self.client,
                                                key) for key, value in self.networks_index.items()}

    def _fetch(self) -> dict:
        return self.client.get_user(self.identifier)



# from __future__ import annotations

# from unittest.mock import MagicMock
import datetime
from typing import Type
from unittest.mock import MagicMock, Mock, patch
from zoneinfo import ZoneInfo
import pytest
import time_machine

import sys
import types

from smarter_client.domain.models import Command, CommandInstance, Commands, Device, LoginSession
from smarter_client.domain.smarter_client import SmarterClient


@pytest.fixture
def SmarterClientMock(mocker):

    mock = MagicMock(
        name='SmarterClient',
        spec={
            'sign_in': Mock(),
            'refresh': Mock(),
            'get_user': Mock(),
            'get_network': Mock(),
            'get_device': Mock(),
            'get_status': Mock(),
            'send_command': Mock(),
            'watch_device_attribute': Mock(),

        }
    )
    return mock


class TestCommands:
    # def from_data_creates_instance(self):
    #     commands: Commands = Commands.from_data(
    #         {'test': {'test-instance': {'value': {'state': 'RCV'}}}})

    #     assert commands.get('test') == isinstance(Command)
    pass

    def test_from_data(self, mocker, SmarterClientMock):
        client = SmarterClientMock()

        assert isinstance(
            Commands.from_data(
                client,
                {'test': {'test-instance': {'value': {'state': 'RCV'}}}},
                mocker.MagicMock()
            ),
            Commands
        )

    def test_commands(self, mocker, SmarterClientMock):
        mock_client = SmarterClientMock()
        mock_device = mocker.MagicMock()

        command_from_data_spy = mocker.spy(Command, 'from_data')
        commands = Commands.from_data(mock_client,
                                      {'test': {
                                          'test-instance': {'value': {'state': 'RCV'}}}},
                                      mock_device
                                      )

        assert isinstance(commands.get('test'), Command)
        command_from_data_spy.assert_called_with(mock_client, {
            'test-instance': {'value': {'state': 'RCV'}}}, 'test', mock_device)

    def test_cannot_be_instantiated_from_id(self, mocker, SmarterClientMock):
        with pytest.raises(RuntimeError):
            Commands.from_id()


class TestCommandInstance:
    def test_from_data(self, mocker, SmarterClientMock):
        mock_client = SmarterClientMock()
        mock_device = mocker.MagicMock()
        mock_command = mocker.MagicMock()
        command = CommandInstance.from_data(
            mock_client,
            {
                'user_id': 'test',
                'value': 1,
                'state': 'RCV',
                'response': 1
            },
            'test',
            mock_command,
            mock_device
        )

        assert command.identifier == 'test'
        assert command.device == mock_device
        assert command.state == 'RCV'


class TestCommand:
    def test_command_from_data_should_create_instance(self, mocker, SmarterClientMock):
        mock_client = SmarterClientMock()
        mock_device = mocker.MagicMock()
        command = Command.from_data(
            mock_client,
            {
                'test-instance': {'value': 0, 'state': 'RCV'}
            },
            'test',
            mock_device
        )

        assert command.identifier == 'test'
        assert command.device == mock_device
        assert command.instances.get('test-instance').state == 'RCV'

    def test_command_execute(self, mocker, SmarterClientMock):
        mock_client = SmarterClientMock()
        mock_device = mocker.MagicMock(identifier='device-1')
        command = Command.from_data(
            mock_client,
            {
                'test-instance': {'value': 0, 'state': 'RCV'}
            },
            'test',
            mock_device
        )

        command.execute('user-1', 5)

        mock_client.send_command.assert_called_with(
            'device-1',
            'test',
            {'value': 5, 'user_id': 'user-1'})


class TestDevice:
    @pytest.fixture
    def device(self, SmarterClientMock):
        mock_client = SmarterClientMock()
        return Device.from_data(
            mock_client,
            {
                'commands': {
                    'test': {
                        'test-instance': {'value': 0, 'state': 'RCV'}
                    }
                },
                'settings': {
                    'network': 'network-1'
                },
                'status': {

                }
            },
            'device-1'
        )

    def test_device_from_data_should_create_instance(self, device):

        assert device.identifier == 'device-1'
        assert device.commands.get('test').identifier == 'test'

    def test_watch_calls_client(self, mocker, device: Device, SmarterClientMock: Type[SmarterClient]):
        mock_client = SmarterClientMock()
        callback = mocker.Mock()
        device.watch(callback)

        mock_client.watch_device_attribute.assert_called_with(
            'device-1', mocker.ANY)

    def test_watch_calls_callback_on_event(self, mocker, device: Device, SmarterClientMock: Type[SmarterClient]):
        mock_client = SmarterClientMock()

        def watch_device_attribute_mock(id, callback):
            callback({'test': 'value'})
            return mocker.Mock()

        mock_client.watch_device_attribute.side_effect = watch_device_attribute_mock
        callback = mocker.Mock()
        device.watch(callback)

        callback.assert_called_with({'test': 'value'})

    def test_watch_twice_raises_error(self, mocker, device: Device, SmarterClientMock: Type[SmarterClient]):
        callback = mocker.Mock()
        device.watch(callback)

        with pytest.raises(RuntimeError):
            device.watch(callback)

    def test_unwatch_does_not_raise_error(self, mocker, device: Device, SmarterClientMock: Type[SmarterClient]):
        callback = mocker.Mock()
        # device.watch(callback)
        device.unwatch()

    def test_unwatch_closes_stream(self, mocker, device: Device, SmarterClientMock: Type[SmarterClient]):
        mock_client = SmarterClientMock()
        callback = mocker.Mock()
        stream_close_mock = mocker.Mock()
        mock_client.watch_device_attribute.return_value = stream_close_mock
        device.watch(callback)

        device.unwatch()

        stream_close_mock.assert_called()

    def test_fetch_calls_client(self, device: Device, SmarterClientMock: Type[SmarterClient]):
        mock_client = SmarterClientMock()
        device.fetch()

        mock_client.get_device.assert_called_with('device-1')


class TestLoginSession:
    def test_login_session_from_data_should_create_instance(self):
        with time_machine.travel(datetime.datetime.fromtimestamp(0, tz=ZoneInfo('UTC')), tick=False):
            session = LoginSession(
                {
                    'kind': 'test-kind',
                    'localId': 'test-id',
                    'email': 'test-email',
                    'displayName': 'test-display-name',
                    'idToken': 'test-token',
                    'registered': True,
                    'refreshToken': 'test-refresh',
                    'expiresIn': 100
                }
            )

            assert session.kind == 'test-kind'
            assert session.local_id == 'test-id'
            assert session.email == 'test-email'
            assert session.display_name == 'test-display-name'
            assert session.id_token == 'test-token'
            assert session.registered == True
            assert session.refresh_token == 'test-refresh'
            assert session.expires_at == datetime.datetime.fromtimestamp(100)

    def test_login_session_is_not_expired(self):
        with time_machine.travel(datetime.datetime.fromtimestamp(0, tz=ZoneInfo('UTC')), tick=False):
            session = LoginSession(
                {
                    'kind': 'test-kind',
                    'localId': 'test-id',
                    'email': 'test-email',
                    'displayName': 'test-display-name',
                    'idToken': 'test-token',
                    'registered': True,
                    'refreshToken': 'test-refresh',
                    'expiresIn': 100
                }
            )

            assert session.is_expired() == False

    def test_login_session_is_expired(self):
        with time_machine.travel(datetime.datetime.fromtimestamp(0, tz=ZoneInfo('UTC')), tick=False) as traveler:
            session = LoginSession(
                {
                    'kind': 'test-kind',
                    'localId': 'test-id',
                    'email': 'test-email',
                    'displayName': 'test-display-name',
                    'idToken': 'test-token',
                    'registered': True,
                    'refreshToken': 'test-refresh',
                    'expiresIn': 100
                }
            )

            traveler.shift(100)
            assert session.is_expired() == True

    def test_login_session_expires_in(self):
        with time_machine.travel(datetime.datetime.fromtimestamp(0, tz=ZoneInfo('UTC')), tick=False) as traveler:
            session = LoginSession(
                {
                    'kind': 'test-kind',
                    'localId': 'test-id',
                    'email': 'test-email',
                    'displayName': 'test-display-name',
                    'idToken': 'test-token',
                    'registered': True,
                    'refreshToken': 'test-refresh',
                    'expiresIn': 100
                }
            )

            assert session.expires_in == 100
            traveler.shift(90)
            assert session.expires_in == 10

    def test_login_session_update(self):
        with time_machine.travel(datetime.datetime.fromtimestamp(0, tz=ZoneInfo('UTC')), tick=False) as traveler:
            session = LoginSession(
                {
                    'kind': 'test-kind',
                    'localId': 'test-id',
                    'email': 'test-email',
                    'displayName': 'test-display-name',
                    'idToken': 'test-token',
                    'registered': True,
                    'refreshToken': 'test-refresh',
                    'expiresIn': 100
                }
            )

            traveler.shift(90)
            session.update({
                'idToken': 'new-token',
                'expiresIn': 101
            })

            assert session.id_token == 'new-token'
            assert session.expires_in == 101
            assert session.expires_at == datetime.datetime.fromtimestamp(191)


class TestNetwork:
    pass


class TestSettings:
    pass


class TestStatus:
    pass


class TestUser:
    pass



from copy import deepcopy
from typing import Generator
import pytest
from smarter_client.dict_util import delete_dict, patch_dict, put_dict


@pytest.fixture()
def original_state() -> Generator[dict, None, None]:
    yield {
        'commands': {
            'test_command': {
                'example': {
                    'state': 'NACK',
                    'value': 0
                }
            }
        },
        'notifications': {
            'user1': {
                'test': False
            }
        },
        'settings': {
            'network': 'test-id',
            'network_ssid': 'test'
        },
        'status': {
            'alarm_failed': 0,
            'alarms': {
                'alarm-id1': {
                    "active": False,
                    "auto_trigger": True,
                    "boil_temperature": 100,
                    "device": "000000000000",
                    "formula_mode_enable": True,
                    "formula_mode_temperature": 70,
                    "formula_mode_temps_locked": True,
                    "hour": 17,
                    "keep_warm_time": 5,
                    "minute": 0,
                    "notify": "Cycle starts",
                    "repeat": "S0M1T0W0T0F0S0",
                    "tone": "SMARTER",
                    "user": "user1",
                    "vendorId": "11111111111"
                }
            },
            "boil_temperature": 81,
            "boiled_keeping_warm": {
                "heating": True,
                "temperature": 100,
                "ts": 1711461225
            },
        }
    }


class TestPatchDict:
    def test_patch_to_empty_dict(self):
        assert patch_dict({}, 'a', {'a1': 1}) == {'a': {'a1': 1}}

    def test_patch_empty_dict_to_dict(self):
        assert patch_dict({'a': {'a1': 0}}, 'a', {}) == {'a': {'a1': 0}}

    def test_patch_empty_dict_to_dict_new_value(self):
        assert patch_dict({'a': {'a1': 0}}, 'b', {}) == {
            'a': {'a1': 0}, 'b': {}}

    def test_patch_dict_to_dict(self, original_state: dict):
        expected = deepcopy(original_state)
        expected['new'] = 1
        assert patch_dict(original_state, '/', {'new': 1}) == expected

    def test_creates_nested(self, original_state: dict):
        expected = deepcopy(original_state)
        expected['commands']['test_command']['test-instance'] = {
            'value': {'state': 'RCV'}
        }

        assert put_dict(
            original_state, 'commands/test_command/test-instance/value', {'state': 'RCV'}) == expected

    def test_deletes_nested(self, original_state: dict):
        expected = deepcopy(original_state)
        original_state['commands']['test_command']['test-instance'] = {
            'value': {'state': 'RCV'}
        }

        assert delete_dict(
            original_state, 'commands/test_command/test-instance') == expected

    def test_delete_root_error(self, original_state: dict):
        with pytest.raises(ValueError):
            delete_dict(original_state, '/')

    def test_put_root(self, original_state: dict):
        assert put_dict(original_state, '/', {'new': 1}) == {'new': 1}

    def test_nested_patch(self):
        assert patch_dict(
            {
                'a1': {
                    'a1b1': 0,
                    'a1b3': 0
                },
                'a2': {
                    'a2b1': 0
                }
            },
            'a1/a1b2',

            {'x1': 1}
        ) == {
            'a1': {
                'a1b1': 0,
                'a1b2': {'x1': 1},
                'a1b3': 0,
            },
            'a2': {
                'a2b1': 0,
            }
        }



import aiohttp
from .._consts import API_KEY


class IdentityToolkitClient:
    def __init__(self, session: aiohttp.ClientSession):
        self.session = session

    async def sign_in(self, email, password):
        request_body = {
            "email": email,
            "password": password,
            "returnSecureToken": True
        }

        async with self.session.post(
                # pylint: disable-next=line-too-long
                f'https://identitytoolkit.googleapis.com/v1/accounts:signInWithPassword?key={API_KEY}',
                data=request_body) as resp:
            assert resp.status == 200
            return await resp.json()



from unittest.mock import MagicMock, Mock, PropertyMock
import pytest

from smarter_client.domain.decorators.session import refreshsession


class TestRefreshSession:
    def test_refreshsession_raises_error_when_no_session(self):
        class Test:
            session = None

            @refreshsession
            def test(self):
                pass  # pragma: no cover

        with pytest.raises(ValueError):
            Test().test()

    def test_refresh_session_when_session_expires_in_less_than_30(self):
        class Test:
            session = type('Session', (), {'expires_in': 29})

            def refresh(self):
                self.session.expires_in = 31

            @refreshsession
            def test(self):
                pass

        Test().test()
        assert Test().session.expires_in == 31

    def test_does_not_refresh_session_when_session_expires_in_more_than_30(self):
        class Test:
            session = type('Session', (), {'expires_in': 31})

            refresh = Mock()

            @refreshsession
            def test(self):
                pass

        test = Test()
        test.test()

        assert test.refresh.call_count == 0
